(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
	typeof define === 'function' && define.amd ? define(['exports'], factory) :
	(factory((global['combinators-p'] = {})));
}(this, (function (exports) { 'use strict';

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
  return typeof obj;
} : function (obj) {
  return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
};



















var defineProperty = function (obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
};





















var slicedToArray = function () {
  function sliceIterator(arr, i) {
    var _arr = [];
    var _n = true;
    var _d = false;
    var _e = undefined;

    try {
      for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
        _arr.push(_s.value);

        if (i && _arr.length === i) break;
      }
    } catch (err) {
      _d = true;
      _e = err;
    } finally {
      try {
        if (!_n && _i["return"]) _i["return"]();
      } finally {
        if (_d) throw _e;
      }
    }

    return _arr;
  }

  return function (arr, i) {
    if (Array.isArray(arr)) {
      return arr;
    } else if (Symbol.iterator in Object(arr)) {
      return sliceIterator(arr, i);
    } else {
      throw new TypeError("Invalid attempt to destructure non-iterable instance");
    }
  };
}();











var toArray = function (arr) {
  return Array.isArray(arr) ? arr : Array.from(arr);
};

var toConsumableArray = function (arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  } else {
    return Array.from(arr);
  }
};

var ncurry = function ncurry(n) {
  var localCurry = function localCurry(name, f) {
    for (var _len = arguments.length, args = Array(_len > 2 ? _len - 2 : 0), _key = 2; _key < _len; _key++) {
      args[_key - 2] = arguments[_key];
    }

    var g = function g() {
      for (var _len2 = arguments.length, largs = Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
        largs[_key2] = arguments[_key2];
      }

      var rest = args.concat(largs);

      if (rest.length < n) return localCurry.apply(undefined, [name, f].concat(toConsumableArray(rest)));
      return f.apply(undefined, toConsumableArray(rest));
    };
    Object.defineProperty(g, "name", { value: name, configureable: true });
    return g;
  };

  return localCurry;
};

var curry2 = ncurry(2);
var curry3 = ncurry(3);
var curry4 = ncurry(4);
var curry5 = ncurry(5);

var isFunction = function isFunction(f) {
  return typeof f === "function";
};
var isThenable = function isThenable(p) {
  return p instanceof Promise || Boolean(p) && ((typeof p === "undefined" ? "undefined" : _typeof(p)) === "object" || typeof p === "function") && isFunction(p.then);
};
// eslint-disable-next-line prefer-destructuring
var isArray = Array.isArray;
var isError = function isError(e) {
  return toString.call(e) === "[object Error]";
};

var ordinal = ["first", "second", "third", "fourth", "fifth"];

var show = function show(p) {
  return JSON.stringify(p, null, 2);
};

var typeError = function typeError(message) {
  throw new TypeError(message);
};

var invalidArgument = function invalidArgument(it, at, expected, actual) {
  return typeError(it + " expects its " + ordinal[at] + " argument to " + expected + "\n  Actual: " + show(actual));
};

var invalidFunction = function invalidFunction(it, at, actual) {
  return invalidArgument(it, at, "be a function", actual);
};

var invalidArray = function invalidArray(it, at, actual) {
  return invalidArgument(it, at, "be an array", actual);
};

var invalidThenable = function invalidThenable(it, at, actual) {
  return invalidArgument(it, at, "be a promise", actual);
};

var of = function of(x) {
  return Promise.resolve(x);
};

var map = curry2("map", function (f, p) {
  if (!isFunction(f)) invalidFunction("Future#map", 0, f);
  if (!isThenable(p)) invalidThenable("Future#map", 1, p);
  return p.then(f);
});

var bimap = curry3("bimap", function (left, right, p) {
  if (!isFunction(left)) invalidFunction("Future#bimap", 0, left);
  if (!isFunction(right)) invalidFunction("Future#bimap", 1, right);
  if (!isThenable(p)) invalidThenable("Future#bimap", 2, p);
  var mark = false; // When set to true, don't run left when right throws.
  return p.then(function (x) {
    mark = true;
    return right(x);
  }).catch(function (e) {
    if (mark) throw e;
    return left(e);
  });
});

var ap = curry2("ap", function (pf, p) {
  if (!isThenable(pf)) invalidThenable("Future#ap", 0, pf);
  if (!isThenable(p)) invalidThenable("Future#ap", 1, p);
  return Promise.all([pf, p]).then(function (_ref) {
    var _ref2 = slicedToArray(_ref, 2),
        f = _ref2[0],
        x = _ref2[1];

    return f(x);
  });
});

var chain = curry2("chain", function (f, p) {
  if (!isFunction(f)) invalidFunction("Future#chain", 0, f);
  if (!isThenable(p)) invalidThenable("Future#chain", 1, p);
  return p.then(f);
});

var Future = { of: of, map: map, bimap: bimap, ap: ap, chain: chain };

var reject = function reject(e) {
  if (isError(e)) return Promise.reject(e);
  return Promise.reject(new Error(e));
};

var clone = function clone(obj) {
  var hash = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : new WeakMap();

  if (Object(obj) !== obj) return obj; // primitives
  if (hash.has(obj)) return hash.get(obj); // cyclic references
  var result = void 0;

  if (Array.isArray(obj)) result = [];else if (obj.constructor) result = new obj.constructor();else result = Object.create(null);

  hash.set(obj, result);

  if (obj instanceof Map) Array.from(obj, function (_ref) {
    var _ref2 = slicedToArray(_ref, 2),
        key = _ref2[0],
        val = _ref2[1];

    return result.set(key, clone(val, hash));
  });
  return Object.assign.apply(Object, [result].concat(toConsumableArray(Object.keys(obj).map(function (key) {
    return defineProperty({}, key, clone(obj[key], hash));
  }))));
};

var tap = curry2("tap", function (f, p) {
  if (!isFunction(f)) invalidFunction("Future#tap", 0, f);
  return of(f(p)).then(function () {
    return p;
  });
});

var tapClone = curry2("tapClone", function (f, p) {
  if (!isFunction(f)) invalidFunction("Future#tapClone", 0, f);
  return of(f(clone(p))).then(function () {
    return p;
  });
});

var constant = (function (x) {
  return function () {
    return of(x);
  };
});

var all = (function (xs) {
  return function () {
    return of(xs).then(function (ys) {
      return Promise.all(ys.map(of));
    });
  };
});

var compose = curry3("compose", function (f, g, x) {
  if (!isFunction(f)) invalidFunction("Future#compose", 0, f);
  if (!isFunction(g)) invalidFunction("Future#compose", 1, g);

  return of(x).then(f).then(g);
});

var flow = curry2("flow", function (fs, x) {
  if (!isArray(fs)) invalidArray("Future#flow", 0, fs);
  return fs.reduce(function (memo, f) {
    return memo.then(f);
  }, of(x));
});

var flow2 = curry3("flow2", function (fs, x, y) {
  if (!isArray(fs)) invalidArray("Future#flow2", 0, fs);

  var _fs = toArray(fs),
      head = _fs[0],
      tail = _fs.slice(1);

  return tail.reduce(function (memo, f) {
    return memo.then(f);
  }, of(head(x, y)));
});

var flow3 = curry4("flow3", function (fs, x, y, z) {
  if (!isArray(fs)) invalidArray("Future#flow3", 0, fs);

  var _fs2 = toArray(fs),
      head = _fs2[0],
      tail = _fs2.slice(1);

  return tail.reduce(function (memo, f) {
    return memo.then(f);
  }, of(head(x, y, z)));
});

var flow4 = curry5("flow4", function (fs, w, x, y, z) {
  if (!isArray(fs)) invalidArray("Future#flow4", 0, fs);

  var _fs3 = toArray(fs),
      head = _fs3[0],
      tail = _fs3.slice(1);

  return tail.reduce(function (memo, f) {
    return memo.then(f);
  }, of(head(w, x, y, z)));
});

var fold = curry3("fold", function (f, acc, xs) {
  if (!isFunction(f)) invalidFunction("Future#fold", 0, f);
  if (!isArray(xs)) invalidArray("Future#fold", 3, xs);
  return xs.reduce(function (memo, x) {
    return memo.then(function (y) {
      return f(y, x);
    });
  }, of(acc));
});

var collectN = function collectN(atOnce, f, xs) {
  var functionName = "Future#collect" + (atOnce > 1 ? atOnce : "");
  if (!isFunction(f)) invalidFunction(functionName, 0, f);
  if (!isArray(xs)) invalidArray(functionName, 1, xs);

  // eslint-disable-next-line promise/avoid-new
  return new Promise(function (resolve, reject) {
    var ret = [];
    var iterator = xs[Symbol.iterator]();
    var isRejected = false;
    var iterableDone = false;
    var resolvingCount = 0;
    var currentIdx = 0;

    var next = function next() {
      if (isRejected) return;
      var nextItem = iterator.next();
      var i = currentIdx;
      currentIdx += 1;

      if (nextItem.done) {
        iterableDone = true;

        if (resolvingCount === 0) resolve(ret);
        return;
      }
      resolvingCount += 1;

      Promise.resolve(nextItem.value).then(f)
      // eslint-disable-next-line promise/always-return
      .then(function (val) {
        ret[i] = val;
        resolvingCount -= 1;
        next(); // eslint-disable-line promise/no-callback-in-promise
      }).catch(function (err) {
        isRejected = true;
        reject(err);
      });
    };

    for (var i = 0; i < atOnce; i += 1) {
      next();
      if (iterableDone) break;
    }
  });
};

var collect = curry3("collect", collectN)(1);
var collect2 = curry3("collect2", collectN)(2);
var collect3 = curry3("collect3", collectN)(3);
var collect4 = curry3("collect4", collectN)(4);
var collect5 = curry3("collect5", collectN)(5);

// Flattens an array of arrays into an array. It only flattens one level deep,
// which is what is required by flatmap.
var flatten = (function (xxs) {
  var result = [];

  if (xxs == null) return result;

  xxs.forEach(function (xs) {
    if (Array.isArray(xs)) result.push.apply(result, toConsumableArray(xs));else result.push(xs);
  });

  return result;
});

var flatmapN = function flatmapN(mapper, f, xs) {
  return mapper(f, xs).then(flatten);
};

var flatmap = curry2("flatmap", function (f, xs) {
  if (!isFunction(f)) invalidFunction("Future#flatmap", 0, f);
  if (!isArray(xs)) invalidArray("Future#flatmap", 1, xs);
  return flatmapN(collect, f, xs);
});

var flatmap2 = curry2("flatmap2", function (f, xs) {
  if (!isFunction(f)) invalidFunction("Future#flatmap2", 0, f);
  if (!isArray(xs)) invalidArray("Future#flatmap2", 1, xs);
  return flatmapN(collect2, f, xs);
});

var flatmap3 = curry2("flatmap3", function (f, xs) {
  if (!isFunction(f)) invalidFunction("Future#flatmap3", 0, f);
  if (!isArray(xs)) invalidArray("Future#flatmap3", 1, xs);
  return flatmapN(collect3, f, xs);
});

var flatmap4 = curry2("flatmap4", function (f, xs) {
  if (!isFunction(f)) invalidFunction("Future#flatmap4", 0, f);
  if (!isArray(xs)) invalidArray("Future#flatmap4", 1, xs);
  return flatmapN(collect4, f, xs);
});

var flatmap5 = curry2("flatmap5", function (f, xs) {
  if (!isFunction(f)) invalidFunction("Future#flatmap5", 0, f);
  if (!isArray(xs)) invalidArray("Future#flatmap5", 1, xs);
  return flatmapN(collect5, f, xs);
});

var caught = curry2("caught", function (f, p) {
  if (!isFunction(f)) invalidFunction("Future#caught", 0, f);
  return of(p).catch(f);
});

var identity = (function (x) {
  return x;
});

var whenElse = curry4("whenElse", function (predicate, consequent, alternative, p) {
  if (!isFunction(predicate)) invalidFunction("Future#whenElse", 0, predicate);
  if (!isFunction(consequent)) invalidFunction("Future#whenElse", 1, consequent);
  if (!isFunction(alternative)) invalidFunction("Future#whenElse", 2, alternative);
  if (!isThenable(p)) invalidThenable("Future#whenElse", 3, p);

  return p.then(function (x) {
    return (
      // eslint-disable-next-line promise/no-nesting
      of(predicate(x)).then(function (bool) {
        return bool === true ? consequent(x) : alternative(x);
      })
    );
  });
});

var unlessElse = curry4("unlessElse", function (predicate, consequent, alternative, p) {
  if (!isFunction(predicate)) invalidFunction("Future#unlessElse", 0, predicate);
  if (!isFunction(consequent)) invalidFunction("Future#unlessElse", 1, consequent);
  if (!isFunction(alternative)) invalidFunction("Future#unlessElse", 2, alternative);
  if (!isThenable(p)) invalidThenable("Future#unlessElse", 3, p);

  return whenElse(predicate, alternative, consequent, p);
});

var when = curry3("when", function (predicate, consequent, p) {
  if (!isFunction(predicate)) invalidFunction("Future#when", 0, predicate);
  if (!isFunction(consequent)) invalidFunction("Future#when", 1, consequent);
  if (!isThenable(p)) invalidThenable("Future#when", 2, p);

  return whenElse(predicate, consequent, identity, p);
});

var unless = curry3("unless", function (predicate, consequent, p) {
  if (!isFunction(predicate)) invalidFunction("Future#unless", 0, predicate);
  if (!isFunction(consequent)) invalidFunction("Future#unless", 1, consequent);
  if (!isThenable(p)) invalidThenable("Future#unless", 2, p);

  return unlessElse(predicate, consequent, identity, p);
});

var lift = function lift(f) {
  for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
    args[_key - 1] = arguments[_key];
  }

  return Promise.all(args).then(function (xs) {
    return f.apply(undefined, toConsumableArray(xs));
  });
};

var lift2 = curry3("lift2", lift);
var lift3 = curry4("lift3", lift);
var lift4 = curry5("lift4", lift);

var spread = curry2("spread", function (f, p) {
  return of(p).then(function (args) {
    if (Array.isArray(args)) return f.apply(undefined, toConsumableArray(args));
    return f(args);
  });
});

var delay = curry2("delay", function (ms, value) {
  var cancel = void 0;

  // eslint-disable-next-line promise/avoid-new
  var thunk = new Promise(function (resolve, reject) {
    var timeOut = setTimeout(function () {
      return resolve(value);
    }, ms);
    cancel = function cancel() {
      if (timeOut) {
        clearTimeout(timeOut);
        timeOut = null;
      }
      reject(new Error("Promise canceled"));
    };
  });
  thunk.cancel = cancel;
  return thunk;
});

var isPromise = isThenable;

Object.defineProperty(isPromise, "name", {
  value: "isPromise",
  configureable: true
});

// times is 0-indexed. A times value of 4, actually makes 5 retries.
var retrier = function retrier(times, waitTime) {
  return function (action) {
    for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      args[_key - 1] = arguments[_key];
    }

    var delayModifier = 1.5;
    var f = isPromise(action) ? function () {
      return action;
    } : action;

    var resolver = function resolver(iter, waitMs) {
      if (iter <= 0) return f.apply(undefined, args);
      return caught(function () {
        return delay(waitMs, null).then(function () {
          return resolver(iter - 1, waitMs * delayModifier);
        });
      }, f.apply(undefined, args));
    };

    return resolver(times, waitTime);
  };
};

var retry = retrier(4, 250);

Object.defineProperty(retry, "name", { value: "retry", configureable: true });

// Using curry allows to fix the number of arguments for functions that are
// implemented as variadic functions.
var retry2 = curry2("retry2", retrier(4, 250));
var retry3 = curry3("retry3", retrier(4, 250));
var retry4 = curry4("retry4", retrier(4, 250));



var api$1 = Object.freeze({
	Future: Future,
	of: of,
	map: map,
	bimap: bimap,
	ap: ap,
	chain: chain,
	ofP: of,
	mapP: map,
	bimapP: bimap,
	apP: ap,
	chainP: chain,
	reject: reject,
	rejectP: reject,
	tap: tap,
	tapClone: tapClone,
	tapP: tap,
	tapCloneP: tapClone,
	constant: constant,
	constantP: constant,
	all: all,
	allP: all,
	compose: compose,
	composeP: compose,
	flow: flow,
	flow2: flow2,
	flow3: flow3,
	flow4: flow4,
	flowP: flow,
	flowP2: flow2,
	flowP3: flow3,
	flowP4: flow4,
	fold: fold,
	foldP: fold,
	collect: collect,
	collect2: collect2,
	collect3: collect3,
	collect4: collect4,
	collect5: collect5,
	collectP: collect,
	collectP2: collect2,
	collectP3: collect3,
	collectP4: collect4,
	collectP5: collect5,
	flatmap: flatmap,
	flatmap2: flatmap2,
	flatmap3: flatmap3,
	flatmap4: flatmap4,
	flatmap5: flatmap5,
	flatmapP: flatmap,
	flatmapP2: flatmap2,
	flatmapP3: flatmap3,
	flatmapP4: flatmap4,
	flatmapP5: flatmap5,
	caught: caught,
	caughtP: caught,
	when: when,
	whenElse: whenElse,
	unless: unless,
	unlessElse: unlessElse,
	whenP: when,
	whenElseP: whenElse,
	unlessP: unless,
	unlessElseP: unlessElse,
	lift2: lift2,
	lift3: lift3,
	lift4: lift4,
	liftP2: lift2,
	liftP3: lift3,
	lift4P: lift4,
	spread: spread,
	spreadP: spread,
	delay: delay,
	delayP: delay,
	retry: retry,
	retry2: retry2,
	retry3: retry3,
	retry4: retry4,
	retryP: retry,
	retryP2: retry2,
	retryP3: retry3,
	retryP4: retry4,
	isPromise: isPromise,
	isPromiseP: isPromise
});

exports['default'] = api$1;
exports.Future = Future;
exports.of = of;
exports.map = map;
exports.bimap = bimap;
exports.ap = ap;
exports.chain = chain;
exports.ofP = of;
exports.mapP = map;
exports.bimapP = bimap;
exports.apP = ap;
exports.chainP = chain;
exports.reject = reject;
exports.rejectP = reject;
exports.tap = tap;
exports.tapClone = tapClone;
exports.tapP = tap;
exports.tapCloneP = tapClone;
exports.constant = constant;
exports.constantP = constant;
exports.all = all;
exports.allP = all;
exports.compose = compose;
exports.composeP = compose;
exports.flow = flow;
exports.flow2 = flow2;
exports.flow3 = flow3;
exports.flow4 = flow4;
exports.flowP = flow;
exports.flowP2 = flow2;
exports.flowP3 = flow3;
exports.flowP4 = flow4;
exports.fold = fold;
exports.foldP = fold;
exports.collect = collect;
exports.collect2 = collect2;
exports.collect3 = collect3;
exports.collect4 = collect4;
exports.collect5 = collect5;
exports.collectP = collect;
exports.collectP2 = collect2;
exports.collectP3 = collect3;
exports.collectP4 = collect4;
exports.collectP5 = collect5;
exports.flatmap = flatmap;
exports.flatmap2 = flatmap2;
exports.flatmap3 = flatmap3;
exports.flatmap4 = flatmap4;
exports.flatmap5 = flatmap5;
exports.flatmapP = flatmap;
exports.flatmapP2 = flatmap2;
exports.flatmapP3 = flatmap3;
exports.flatmapP4 = flatmap4;
exports.flatmapP5 = flatmap5;
exports.caught = caught;
exports.caughtP = caught;
exports.when = when;
exports.whenElse = whenElse;
exports.unless = unless;
exports.unlessElse = unlessElse;
exports.whenP = when;
exports.whenElseP = whenElse;
exports.unlessP = unless;
exports.unlessElseP = unlessElse;
exports.lift2 = lift2;
exports.lift3 = lift3;
exports.lift4 = lift4;
exports.liftP2 = lift2;
exports.liftP3 = lift3;
exports.lift4P = lift4;
exports.spread = spread;
exports.spreadP = spread;
exports.delay = delay;
exports.delayP = delay;
exports.retry = retry;
exports.retry2 = retry2;
exports.retry3 = retry3;
exports.retry4 = retry4;
exports.retryP = retry;
exports.retryP2 = retry2;
exports.retryP3 = retry3;
exports.retryP4 = retry4;
exports.isPromise = isPromise;
exports.isPromiseP = isPromise;

Object.defineProperty(exports, '__esModule', { value: true });

})));
//# sourceMappingURL=combinators-p.js.map
